;
;  PROJECT:         Advanced DLL Wrapper
;  FILE:            inc\panalyze.asm
;  PURPOSE:         function parameters counter
;                   (recursive code analyzer)
;  PROGRAMMER:      90210
;
;
;  15/05/2002 23:20:34 90210
;
;  TODO:
;       - SEH
;       - check branches' starts for residing in module
;       - correct opcode prefixes handling
;       - more sophisticated code analyzing via import jumps checking
;       - store regions of analyzed code, not just jump's destinations
;
;  HISTORY:         15/05/2002: Created
;

if DEBUG_BUILD
extrn                   OutputDebugStringA: proc
extrn                   wsprintfA         : proc
endif

.data

ods          macro      string
if DEBUG_BUILD
             call       OutputDebugStringA,offset string
endif
endm

if DEBUG_BUILD
entering     db         'PANALYZER: Entering %08X',0
exiting1     db         'PANALYZER: Exiting %08X: both branches analyzed',0
exiting2     db         'PANALYZER: Exiting %08X: RET %04X (%6X) found at %08X',0
exiting3     db         'PANALYZER: Exiting %08X: unpredictable jump at %8X',0
;badopcode    db         'PANALYZER: Bad opcode found at %08X',0
branches     db         'PANALYZER: Branch: %08X (conditional jump at %08X) -> %08X',0
jmpfound     db         'PANALYZER: JMP %08X found at %08X (branch %08X)',0
wasthere     db         'PANALYZER: JMP %08X ignored: already analyzed',0
endif

tbl          db         2048 dup (?)
retdata      dd         1000 dup (?)
eiptable     dd         100000 dup (?)


if DEBUG_BUILD
debugbuff    db         100 dup (?)
endif

.code
analyzeparams proc      starteip:dword,stackmem:dword
             uses       ebx,esi,edi
             local      oldstack:dword

             lea        ebx,tbl
             call       disasm_init,ebx
             add        esp,4

             and        dword ptr retdata,0
             and        dword ptr eiptable,0

;             mov        oldstack,esp
;             mov        eax,stackmem
;             lea        esp,[eax+STACK_SIZE-10000]
             call       panalyzer,starteip
;             mov        esp,oldstack
             ret

analyzeparams endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  FUNCTION: adddata
;;                checks if supplied dword has been already placed
;;                into the table; places it there if it hasn't.
;;
;;  PARAMETERS:
;;    table   :   table of some data
;;    _data   :   dword to check and place
;;
;;  RETURNS:
;;    eax     :   FALSE if not found, TRUE otherwise
adddata      proc       table:dword,_data:dword
             uses       esi,ecx
             mov        ecx,_data
             mov        esi,table
trynexteip:  lodsd
             or         eax,eax
             jz         alleipstried
             cmp        eax,ecx
             jnz        trynexteip
             jmp        _zeend
alleipstried:mov        [esi-4],ecx
             and        dword ptr [esi],0
_zeend:      ret
adddata      endp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  FUNCTION: calcdest32
;;                calculates eip after 5- or 6-bytes jump
;;  PARAMETERS:
;;    esi     :   current eip
;;    ecx     :   opcode size (instruction without operand)
;;
;;  RETURNS:
;;    eax     :   eip after jump
calcdest32:  mov        eax,[esi+ecx]
             lea        eax,[eax+ecx+4]
             add        eax,esi
             ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  FUNCTION: calcdest8
;;                calculates eip after 2 bytes jump
;;  PARAMETERS:
;;    esi     :   current eip
;;
;;  RETURNS:
;;    eax     :   eip after jump
calcdest8:   mov        al,byte ptr [esi+1]
             movsx      eax,al           ;set al sign to whole register
             lea        eax,[esi+eax+2]
             ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  FUNCTION: shouldbreak
;;                checks if current opcode transfers execution to
;;                undeterminable place
;;  PARAMETERS:
;;    esi     :   opcode eip
;;
;;  RETURNS:
;;    eax     :   FALSE if normal opcode, TRUE otherwise
shouldbreak:
;bad jumps: 0ff20h-0ff2fh, 0ff60h-0ff6fh, 0ffe0h-0ffefh
             movzx      eax,word ptr [esi]
             cmp        al,0ffh
             jnz        goodopcode
             cmp        ah,20h
             jb         goodopcode
             cmp        ah,30h
             jb         badopcode

             cmp        ah,60h
             jb         goodopcode
             cmp        ah,70h
             jb         badopcode

             cmp        ah,0e0h
             jb         goodopcode
             cmp        ah,0f0h
             jb         badopcode

goodopcode:  xor        eax,eax
badopcode:   ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  FUNCTION: panalyzer
;;                analyzes given procedure and finds number of
;;                its parameters
;;  PARAMETERS:
;;    _eip    :   eip to start from
;;
;;  RETURNS:
;;    eax     :   found number of parameters

panalyzer    proc       _eip:dword
             uses       ebx,esi,edi,edx,ebp

             mov        esi,_eip

if DEBUG_BUILD
             call       wsprintfA,offset debugbuff,offset entering,esi
             add        esp,4*3
             ods        debugbuff
endif

analyze_next:
             call       shouldbreak   ;esi=current analyzer eip
             or         eax,eax
             jz         normalopcode

if DEBUG_BUILD                        ;can't calc jump destination point
             call       wsprintfA,offset debugbuff,offset exiting3,_eip,esi
             add        esp,4*4
             ods        debugbuff
endif
             ret

normalopcode:mov        al,byte ptr [esi]

             cmp        al,0c3h
             jnz        check_retX
;FIXME FIXME FIXME
             call       adddata,offset retdata,0ffh  ;simple ret
if DEBUG_BUILD
             xor        eax,eax
             mov        al,0c3h
             call       wsprintfA,offset debugbuff,offset exiting2,_eip,0,eax,esi
             add        esp,4*6
             ods        debugbuff
endif

             jmp        _2ret
check_retX:  cmp        al,0c2h
             jnz        other_opcode
             movzx      eax,word ptr [esi+1]
             call       adddata,offset retdata,eax
if DEBUG_BUILD
             pusha
             xor        eax,eax
             mov        al,0c2h
             shl        eax,16
             movzx      ebx,word ptr [esi+1]
             mov        ax,bx
             xchg       al,ah
             call       wsprintfA,offset debugbuff,offset exiting2,_eip,ebx,eax,esi
             add        esp,4*6
             ods        debugbuff
             popa
endif

_2ret:       ret
other_opcode:
;FIXME FIXME FIXME


             cmp        al,0fh
             jb         next_opcode
             jz         _0f

             cmp        al,70h        ;check for simple jxx
             jb         next_opcode   ;
             cmp        al,80h        ;
             jb         _condjmps     ;

             cmp        al,0e0h       ;check for ecx-oriented
             jb         next_opcode   ;
             cmp        al,0e3h       ;
             jb         _condjmps     ;

             cmp        al,0e9h       ;5-bytes jmp
             jnz        check2bytes
             push       1
             pop        ecx
             call       calcdest32
             jmp        its_jmp


check2bytes: cmp        al,0ebh       ;2-bytes jmp
             jnz        next_opcode
             call       calcdest8
its_jmp:

if DEBUG_BUILD
             pusha
             call       wsprintfA,offset debugbuff,offset jmpfound,eax,esi,_eip
             add        esp,4*5
             ods        debugbuff
             popa
endif
             mov        ecx,eax
             call       adddata,offset eiptable,eax
             or         eax,eax
             jz         wasntthere

if DEBUG_BUILD
             pusha
             call       wsprintfA,offset debugbuff,offset wasthere,ecx
             add        esp,4*3
             ods        debugbuff
             popa
endif
             ret                       ;jump destination already analyzed

wasntthere:  mov        esi,ecx
             jmp        analyze_next

_0f:         push       ebp
             mov        al,byte ptr [esi+1]
             cmp        al,080h        ;check for extended jumps (0f ...)
             jb         next_opcode    ;
             cmp        al,08fh        ;
             ja         next_opcode    ;

             push       2
             pop        ecx            ; 0f 8x [4-byte operand]
             call       calcdest32
             push       6
             pop        ebp
             jmp        makebranches

_condjmps:   push       ebp
             call       calcdest8
             push       2
             pop        ebp
makebranches:mov        edi,eax

if DEBUG_BUILD
             mov        eax,ebp
             pop        ebp
             push       ebp

             mov        ecx,_eip
             mov        ebp,eax
             pusha
             call       wsprintfA,offset debugbuff,offset branches,ecx,esi,edi
             add        esp,4*5
             ods        debugbuff
             popa
endif
             lea        edx,eiptable
             call       adddata,edx,edi ;edi=jump destination
             or         eax,eax
             jnz        check_2nd_branch

             call       panalyzer,edi
check_2nd_branch:
             lea        ecx,[esi+ebp]          ;ecx=eip after jump opcode
if DEBUG_BUILD
             mov        eax,ebp
             pop        ebp
             push       ebp

             push       eax
             mov        eax,_eip
             pop        ebp
             pusha
             call       wsprintfA,offset debugbuff,offset branches,eax,esi,ecx
             add        esp,4*5
             ods        debugbuff
             popa
endif
             call       adddata,edx,ecx
             or         eax,eax
             jnz        all_done_here
             call       panalyzer,ecx
all_done_here:                        ;both branches already analyzed
             pop        ebp

if DEBUG_BUILD
             mov        esi,_eip
             call       wsprintfA,offset debugbuff,offset exiting1,esi
             add        esp,4*3
             ods        debugbuff
endif
             ret

next_opcode: call       disasm_main,ebx,esi
             add        esp,4*2
             add        esi,eax
             jmp        analyze_next
             ret
panalyzer    endp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

_2analyze:   jnz        gg
             nop
ff:          nop
             nop
gg:          mov        eax,2
             jmp ff
             ret

;return opcodes

;             db         0c2h     ;   ret       45105
;             dw         45105
;             db         0c3h     ;   ret
;             db         0cah     ;   retf      45105   \
;             dw         45105                           > not used in win32
;             db         0cbh     ;   retf              /
;total: 0c2h-0c3h


;inconditional jumps (0e9h-0ebh)

;             db         0e9h     ;   jmp        $+90210h
;             dd         90210h-5
;             db         0eah     ;   jmp        1234h:12345678h
;             dd         12345678h               not used in win32
;             dw         1234h
;             dw         0ebh     ;   jmp        $+2
;total: 0e9h-0ebh

;conditional jumps

;             dw         70h      ;   jo         $+2
;             dw         71h      ;   jno        $+2
;             dw         72h      ;   jb         $+2
;             dw         73h      ;   jnb        $+2
;             dw         74h      ;   jz         $+2
;             dw         75h      ;   jnz        $+2
;             dw         76h      ;   jbe        $+2
;             dw         77h      ;   jnbe       $+2
;             dw         78h      ;   js         $+2
;             dw         79h      ;   jns        $+2
;             dw         7ah      ;   jp         $+2
;             dw         7bh      ;   jnp        $+2
;             dw         7ch      ;   jl         $+2
;             dw         7dh      ;   jnl        $+2
;             dw         7eh      ;   jng        $+2
;             dw         7fh      ;   jg         $+2
;total: 070h-07fh

;ecx oriented jumps

;             dw         0e0h     ;   loopnz     $+2
;             dw         0e1h     ;   loopz      $+2
;             dw         0e2h     ;   loop       $+2
;             dw         0e3h     ;   jecxz      $+2
;total: 0e0h-0e3h

;extended contitionals

;             dw         800fh    ;   jo         $+5
;             dd         0
;             dw         810fh    ;   jno        $+5
;             dd         0
;             dw         820fh    ;   jb         $+5
;             dd         0
;             dw         830fh    ;   jnb        $+5
;             dd         0
;             dw         840fh    ;   jz         $+5
;             dd         0
;             dw         850fh    ;   jnz        $+5
;             dd         0
;             dw         860fh    ;   jbe        $+5
;             dd         0
;             dw         870fh    ;   jnbe       $+5
;             dd         0
;             dw         880fh    ;   js         $+5
;             dd         0
;             dw         890fh    ;   jns        $+5
;             dd         0
;             dw         8a0fh    ;   jp         $+5
;             dd         0
;             dw         8b0fh    ;   jnp        $+5
;             dd         0
;             dw         8c0fh    ;   jl         $+5
;             dd         0
;             dw         8d0fh    ;   jnl        $+5
;             dd         0
;             dw         8e0fh    ;   jng        $+5
;             dd         0
;             dw         8f0fh    ;   jg         $+5
;             dd         0
;total: 0f80h-0f8fh

;break opcodes

;             db         0FFh,0E0h           ; jmp eax
;             db         0FFh,0E1h           ; jmp ecx
;             db         0FFh,0E2h           ; jmp edx
;             db         0FFh,0E3h           ; jmp ebx
;             db         0FFh,0E4h           ; jmp esp
;             db         0FFh,0E5h           ; jmp ebp
;             db         0FFh,0E6h           ; jmp esi
;             db         0FFh,0E7h           ; jmp edi

;             db         0FFh,0E8h           ; jmp ax
;             db         0FFh,0E9h           ; jmp cx
;             db         0FFh,0EAh           ; jmp dx
;             db         0FFh,0EBh           ; jmp bx
;             db         0FFh,0ECh           ; jmp sp
;             db         0FFh,0EDh           ; jmp bp
;             db         0FFh,0EEh           ; jmp si
;             db         0FFh,0EFh           ; jmp di
;total: 0ffe0h-0ffefh

;             db         0FFh,20h            ; jmp dword ptr [eax]
;             db         0FFh,21h            ; jmp dword ptr [ecx]
;             db         0FFh,22h            ; jmp dword ptr [edx]
;             db         0FFh,23h            ; jmp dword ptr [ebx]
;             db         0FFh,24h,0Dh        ; jmp dword ptr [00000001+ecx]
;             dd         00000001
;             db         0FFh,25h            ; jmp dword ptr [00000000]
;             dd         00000000
;             db         0FFh,26h            ; jmp dword ptr [esi]
;             db         0FFh,27h            ; jmp dword ptr [edi]

;             db         0FFh,28h            ; jmp pword ptr [eax]
;             db         0FFh,29h            ; jmp pword ptr [ecx]
;             db         0FFh,2Ah            ; jmp pword ptr [edx]
;             db         0FFh,2Bh            ; jmp pword ptr [ebx]
;             db         0FFh,2Ch,0Dh        ; jmp pword ptr [00000001+ecx]
;             dd         00000001
;             db         0FFh,2Dh            ; jmp pword ptr [00000000]
;             dd         00000000
;             db         0FFh,2Eh            ; jmp pword ptr [esi]
;             db         0FFh,2Fh            ; jmp pword ptr [edi]
;total: 0ff20h-0ff2fh

;             db         0FFh,60h,00         ; jmp dword ptr [eax]
;             db         0FFh,61h,00         ; jmp dword ptr [ecx]
;             db         0FFh,62h,00         ; jmp dword ptr [edx]
;             db         0FFh,63h,00         ; jmp dword ptr [ebx]
;             db         0FFh,64h,00,00      ; jmp dword ptr [eax+eax]
;             db         0FFh,65h,00         ; jmp dword ptr [ebp]
;             db         0FFh,66h,00         ; jmp dword ptr [esi]
;             db         0FFh,67h,00         ; jmp dword ptr [edi]

;             db         0FFh,68h,00         ; jmp pword ptr [eax]
;             db         0FFh,69h,00         ; jmp pword ptr [ecx]
;             db         0FFh,6Ah,00         ; jmp pword ptr [edx]
;             db         0FFh,6Bh,00         ; jmp pword ptr [ebx]
;             db         0FFh,6Ch,00,00      ; jmp pword ptr [eax+eax]
;             db         0FFh,6Dh,00         ; jmp pword ptr [ebp]
;             db         0FFh,6Eh,00         ; jmp pword ptr [esi]
;             db         0FFh,6Fh,00         ; jmp pword ptr [edi]
;total: 0ff60h-0ff6fh

include       incs\lde32.inc

